/******************************************************************************
 *
 * Copyright 2014 Altera Corporation. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of the author may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 ******************************************************************************/
 
/********************************************************************************//*!
 * @file  minimal_io.c
 * @brief  Minimal IO support utility C source code file.
 * 
 * @details  Minimal IO support utility C source code file.
 * 
 * @note  nothing. 
 * 
 * @attention 
 * Copyright (C) 2013-2019 MACNICA,Inc. All Rights Reserved.\n
 *   This software is licensed "AS IS". 
 *   Please perform use of this software by a user's own responsibility and expense.
 *   It cannot guarantee in the maker side about the damage which occurred by the ab-
 *   ility not to use or use this software, and all damage that occurred secondarily.
 **//*******************************************************************************/

/*
 * This file contains minimal IO support to make printf work over UART
 * The code relies on the Preloader to set the UART module including baud rate
 */

#include <stdio.h>
#include <inttypes.h>
#include <unistd.h>

#include "socal/alt_uart.h"
#include "socal/hps.h"
#include "socal/socal.h"

//Select Line Feed Code
#define CR_LF_MODE 1

//Please Select UART ch for your system
#ifndef UART_CH
#define UART_CH (0) // 0 or 1
#endif
#if UART_CH
#define UART_LSR_ADDR         (ALT_UART1_LSR_ADDR)
#define UART_RBR_THR_DLL_ADDR (ALT_UART1_RBR_THR_DLL_ADDR)
#else
#define UART_LSR_ADDR         (ALT_UART0_LSR_ADDR)
#define UART_RBR_THR_DLL_ADDR (ALT_UART0_RBR_THR_DLL_ADDR)
#endif


int _close(int file)
{
    /* Succeeds only for STDOUT */
    return (file == STDOUT_FILENO) ? 0 : -1;
}

int _fstat(int file, void *st)
{
    /* Succeeds only for STDOUT */
    return (file == STDOUT_FILENO) ? 0 : -1;
}

int _isatty(int file)
{
    /* Succeeds only for STDOUT */
    return (file == STDOUT_FILENO) ? 1 : -1;
}

off_t _lseek(int file, off_t ptr, int dir)
{
    /* Succeeds only for STDOUT */
    return (file == STDOUT_FILENO) ? 0 : -1;
}

//TODO: MACNICA Mod Start
#if 0
int _read(int file, void *ptr, size_t len)
{
    /* Always fails */
    return -1;
}
#else
int _read(int file, char *ptr, size_t len)
{
	char ch;
	
	/* Fails if not STDOUT */
	if(file != STDIN_FILENO)
	{
		return -1;
	}

	/* Print each character to UART */
	for(int index=0; index<len; )
	{
		/* Wait until DR is ready */
		while(1 != ALT_UART_LSR_DR_GET(alt_read_word(UART_LSR_ADDR)));
		
		/* Read character from THR */
		ch = alt_read_byte(UART_RBR_THR_DLL_ADDR);
		
		if(ch == '\b'){
			if(index > 0){
				ptr[index--] = '\0';
		    	/* Echo back */
		    	while(1 != ALT_UART_LSR_THRE_GET(alt_read_word(UART_LSR_ADDR)));
				alt_write_word(UART_RBR_THR_DLL_ADDR, '\b');
		    	while(1 != ALT_UART_LSR_THRE_GET(alt_read_word(UART_LSR_ADDR)));
				alt_write_word(UART_RBR_THR_DLL_ADDR, ' ');
		    	while(1 != ALT_UART_LSR_THRE_GET(alt_read_word(UART_LSR_ADDR)));
				alt_write_word(UART_RBR_THR_DLL_ADDR, '\b');
			}
		} else if((ch == '\r')||(ch == '\n')){
			if(ch == '\r'){
				ptr[index++] = ch;
				ch = '\n';
				/* Echo back */
				while(1 != ALT_UART_LSR_THRE_GET(alt_read_word(UART_LSR_ADDR)));
				alt_write_word(UART_RBR_THR_DLL_ADDR, '\r');
			}
			ptr[index++] = ch;
			/* Echo back */
			while(1 != ALT_UART_LSR_THRE_GET(alt_read_word(UART_LSR_ADDR)));
			alt_write_word(UART_RBR_THR_DLL_ADDR, '\n');

			return index;		/* input finish */

		} else {
			if(ch == '\x03'){
				ptr[0] = ch;
				ptr[1] = '\n';
				ptr[2] = '\0';

				/* Echo back */
				while(1 != ALT_UART_LSR_THRE_GET(alt_read_word(UART_LSR_ADDR)));
				alt_write_word(UART_RBR_THR_DLL_ADDR, '\r');
				while(1 != ALT_UART_LSR_THRE_GET(alt_read_word(UART_LSR_ADDR)));
				alt_write_word(UART_RBR_THR_DLL_ADDR, '\n');

				return 2;		/* input cancel */

			} else if((ch < ' ')||('~' < ch)){
				ch = ' ';
			}
			ptr[index++] = ch;
			/* Echo back */
			while(1 != ALT_UART_LSR_THRE_GET(alt_read_word(UART_LSR_ADDR)));
			alt_write_word(UART_RBR_THR_DLL_ADDR, ch);
		}
	}
	
	return len;
}
#endif
//TODO: MACNICA Mod End

int _write(int file, char * ptr, unsigned len, int flag )
{
    /* Fails if not STDOUT */
    if(file != STDOUT_FILENO)
    {
        return -1;
    }
#if CR_LF_MODE
    /* Print each character to UART */
    for(int i=0; i<len; i++)
    {
        /* Wait until THR is empty */
        while(1 != ALT_UART_LSR_THRE_GET(alt_read_word(UART_LSR_ADDR)));

        /* Insert '\r' if detect '\r' */
        if(ptr[i] == '\n'){
        	alt_write_word(UART_RBR_THR_DLL_ADDR, '\r');
        	/* Wait until THR is empty */
        	while(1 != ALT_UART_LSR_THRE_GET(alt_read_word(UART_LSR_ADDR)));
        }

        /* Write character to THR */
        alt_write_word(UART_RBR_THR_DLL_ADDR, ptr[i]);
    }
#else
    /* Print each character to UART */
    for(int i=0; i<len; i++)
    {
        /* Wait until THR is empty */
        while(1 != ALT_UART_LSR_THRE_GET(alt_read_word(UART_LSR_ADDR)));

        /* Write character to THR */
        alt_write_word(UART_RBR_THR_DLL_ADDR, ptr[i]);
    }
#endif

    /* All printed fine */
    return len;
}

